from collections import defaultdict

from .table import Table
from .terms import Term
from .tools import term


class ProjectionSet:
    def __init__(self, table, usls):
        super().__init__()
        self.table = table

        self.terms = defaultdict(list)
        self.projection = defaultdict(list)

        for u in usls:
            for t in u.objects(Term).intersection([cell.term for cell in self.table]):
                cell = self.table[t]
                self.terms[cell].append(u)
                self.projection[u].append(cell)

    @property
    def ratio(self):
        return len(self.projection) * len(self.terms) / len(self.table)

if __name__ == '__main__':
    from ieml.usl import random_usl
    from ieml.usl.template import Template

    usls = list(Template("[([O:M:.]+[x.M:M:.-])]", ['r0', 'r1']))

    paradigms = [t for t in term("M:O:.M:M:.-").relations.contains if len(t) != 1 and t.ntable == 1]

    projections = [ProjectionSet(Table(p), usls) for p in paradigms]

    projections= sorted(projections, key=lambda p: p.ratio, reverse=True)

    proj = ProjectionSet(Table("M:O:.M:M:.-"), usls)
    print(proj)


import logging
import os
import ply.yacc as yacc

from ieml import parser_folder
from ieml.exceptions import CannotParse
from ieml.syntax.parser import IEMLParser
from ieml.usl.parser.lexer import tokens, get_lexer
from ieml.usl import Usl
from ieml.metaclasses import Singleton


class USLParser(metaclass=Singleton):
    tokens = tokens

    def __init__(self):

        # Build the lexer and parser
        self.lexer = get_lexer()
        self.parser = yacc.yacc(module=self, errorlog=logging, start='usl',
                                debug=False, optimize=True, picklefile=os.path.join(parser_folder, "usl_parser.pickle"))

    def parse(self, s):
        """Parses the input string, and returns a reference to the created AST's root"""
        self.usl = s
        self.root = None
        self.parser.parse(s, lexer=self.lexer)

        if self.root is not None:
            return self.root
        else:
            raise CannotParse(s)

    # Parsing rules
    def p_usl(self, p):
        """ usl : IEML_OBJECT """
        self.root = Usl(IEMLParser().parse(p[1]))

    def p_error(self, p):
        if p:
            print("Syntax error at '%s' (%d, %d)" % (p.value, p.lineno, p.lexpos))
        else:
            print("Syntax error at EOF")

        raise CannotParse(self.usl)

if __name__ == '__main__':
    print(str(USLParser().parse('{[wa.]}')))



from ieml.dictionary.terms import Term
from ieml.syntax.parser.parser import IEMLParser
from ieml.syntax.words import Word, Morpheme


def proposition(arg, dictionary=None):
    if isinstance(arg, str):
        arg = IEMLParser().parse(arg)

    if isinstance(arg, Term):
        return Word(root=Morpheme([arg]))

    return arg


import copy
import itertools
import numpy as np

from ..constants import LANGUAGES
from ..dictionary import Term, term
from ..syntax import Word
from .paths import path
from .tools import usl, replace_paths


class Template:
    def __init__(self, model, path_list):
        """

        :param model:
        :param path_list:
        :param translation_rule:
        """
        super().__init__()

        self.model = usl(model)

        if not isinstance(self.model.ieml_object, Word):
            raise ValueError("Template only implemented for words.")

        self.paths = [path(p) for p in path_list]
        if any(p.cardinal != 1 for p in self.paths):
            raise ValueError("Can only build a template from singular path (no '+')")

        self.multiples = []
        self.result = []

        self.build()

    def build(self):
        self.multiples = []
        for i, p in enumerate(self.paths):
            t = self.model[p]
            if not isinstance(t, Term) or t.script.cardinal == 1:
                raise ValueError("Invalid path for template creation [%s]->'%s' leading to a non term object,"
                                 "or the term is not a paradigm."%(str(p), str(t)))

            self.multiples.append({
                'index': i,
                'path': p,
                'term': t,
            })

        self.template = np.zeros(shape=tuple(t['term'].script.cardinal for t in self.multiples), dtype=object)

        for index in itertools.product(*tuple(range(s) for s in self.template.shape)):
            self.template[index] = replace_paths(self.model, {
                m['path']: term(m['term'].script.singular_sequences[index[i]]) for i, m in enumerate(self.multiples)
            })

    def get_translations(self, translation_template, mapping=None):
        """

        :param translation_template: str, a string with special string embedded to determine where to put the variation
        translations
        :param mapping: dict {path -> str} where str is the special string in translation template to replace with the
        term translation
        :return: dict {usl -> dict { str (language) ->  str}}
        """

        if mapping is None:
            mapping = {m['path']: str(m['path']) for m in self.multiples}

        def _replace(u):
            translation = copy.deepcopy(translation_template)
            for m in self.multiples:
                for l in LANGUAGES:
                    translation[l] = translation[l].replace(mapping[m['path']], u[m['path']].translations[l])
            return translation

        return {u: _replace(u) for u in self.__iter__()}


    def __iter__(self):
        return (np.asscalar(w) for w in np.nditer(self.template, flags=['refs_ok']))